#include <stdio.h>
#include <stdlib.h>
#include <stm32f10x.h>
#include "usbmidi.h"

#ifdef DEBUG
  #include "debugio.h"
#else
  #define debug_printf(...)
#endif

void ClockConfig()
{
  // Resetvalue laut Datenblatt
  RCC->CR=0x81;
  
  // warten bis HSI an
  while ((RCC->CR & RCC_CR_HSIRDY)==0)  
    ;

  // externen 8 MHz Oszillator HSE einschalten 
  RCC->CR |= RCC_CR_HSEON;       

  while ((RCC->CR & RCC_CR_HSERDY)==0)  
    ;

  // Prefetchbuffer und 2 Waitstates
  FLASH->ACR=FLASH_ACR_PRFTBE | FLASH_ACR_LATENCY_2;

  RCC->CFGR=  RCC_CFGR_HPRE_DIV1    // 
            | RCC_CFGR_PPRE1_DIV2   // ABP1 36MHz
            | RCC_CFGR_PPRE2_DIV1   // ABP2 72MHz
            | RCC_CFGR_ADCPRE_DIV8  // ADC prescaler
            | RCC_CFGR_PLLMULL9     // PLL * 9
            | RCC_CFGR_PLLSRC_HSE;  // HSE fuer PLL

  // PLL einschalten
  RCC->CR |= RCC_CR_PLLON;
  while ((RCC->CR & RCC_CR_PLLRDY)==0)
    ;

  // PLL als Taktquelle
  RCC->CFGR |= RCC_CFGR_SW_PLL;
  while ((RCC->CFGR & 0b1100) != RCC_CFGR_SWS_PLL)
    ;
}

enum eGpioInOutMode
{
  GPIO_IN_ANALOG        = 0x00,
  GPIO_IN_FLOAT         = 0x01,
  GPIO_IN_PULLUPDOWN    = 0x02,

  GPIO_OUT_NORMAL       = 0x00,
  GPIO_OUT_OPENDRAIN    = 0x01,
  GPIO_OUT_ALTNORMAL    = 0x02,
  GPIO_OUT_ALTOPENDRAIN = 0x03
};

enum eGpioMode
{
  GPIO_IN               = 0x00,
  GPIO_OUT_10MHZ        = 0x01,
  GPIO_OUT_2MHZ         = 0x02,
  GPIO_OUT_50MHZ        = 0x03
};

enum eGpioPo
{
  PA = GPIOA_BASE,
  PB = GPIOB_BASE,
  PC = GPIOC_BASE,
  PD = GPIOD_BASE,
};  

enum eGpioState
{
  GPIO_INIT,
  GPIO_SET,
  GPIO_CLR,
  GPIO_UNINIT
};  

template<eGpioPo ePort, uint32_t bitPosi, eGpioMode mode=GPIO_IN, eGpioInOutMode inoutmode=GPIO_IN_ANALOG>
class GpioPin 
{
  // inline GPIO_TypeDef *GetPort() { return (GPIO_TypeDef *)ePort; }
  constexpr GPIO_TypeDef * GetPort() { return (reinterpret_cast<GPIO_TypeDef*>(ePort)); }   
public:
  inline  GpioPin(eGpioState eState=GPIO_INIT)    
    { 
    switch (eState)
      {
      case GPIO_INIT: { Init(); return; }
      case GPIO_SET:  { Init(); Set(); return; }
      case GPIO_CLR:  { Init(); Clr(); return; }
      default:        { return; }
      }
    }
  
  inline void Init()            
    {
    uint32_t m=(inoutmode&0x03)<<2;
    m|=(mode&0x03);
    RCC->APB2ENR |= RCC_APB2ENR_AFIOEN; 
    switch (ePort)
      {
      case PA: RCC->APB2ENR |= RCC_APB2ENR_IOPAEN;  break;
      case PB: RCC->APB2ENR |= RCC_APB2ENR_IOPBEN;  break;
      case PC: RCC->APB2ENR |= RCC_APB2ENR_IOPCEN;  break;
      case PD: RCC->APB2ENR |= RCC_APB2ENR_IOPDEN;  break;
      }

    if (bitPosi >=8)
      {
      uint32_t r=GetPort()->CRH;
      r&= ~(0b1111<<(((bitPosi-8)&0x7)*4));
      r|= ((m & 0b1111)<<(((bitPosi-8)&0x7)*4));
      GetPort()->CRH=r;
      }
    else
      {
      uint32_t r=GetPort()->CRL;
      r&= ~(0b1111<<((bitPosi&0x7)*4));
      r|= ((m & 0b1111)<<((bitPosi&0x7)*4));
      GetPort()->CRL=r;
      }
    }
   
  inline void Set()             { GetPort()->BSRR  = (1<<bitPosi); }
  inline void Clr()             { GetPort()->BRR   = (1<<bitPosi); }
  inline void Toggle()          { (((GetPort()->ODR & (1<<bitPosi))) !=0) ? Clr() : Set(); }
  inline void SetValue(bool v)  { v ? Set() : Clr();   } 
  inline bool IsSet()           { return (GetPort()->ODR & (1<<bitPosi)) !=0; }
  inline bool In()              { return (GetPort()->IDR & (1<<bitPosi)) !=0; }
};

// Usb-Pin mit 1.5k Widerstand auf GND ziehen
// das passiert vor main() ! 
GpioPin<PA, 12, GPIO_OUT_2MHZ, GPIO_OUT_OPENDRAIN> aA12;

int main(void) 
{
  ClockConfig();

  // OpenDrain von aA12 auf High setzen
  // damit die Enumeration startet
  aA12.Set();
  UsbSetup();

  while (1)
    {
    uint32_t i=UsbGetMidiMsg(); 
    if (i>0)
      {
      debug_printf("%08x\n", i);
      UsbMidiMsgOut(i);
      }
    }

  return 0;
}

